from norminette.lexer import Lexer
from norminette.rules import Rule
from norminette.scope import GlobalScope


class CheckPreprocessorInclude(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsPreprocessorStatement"]

    def run(self, context):
        """
        Includes must be at the start of the file
        You cannot include anything that isn't an header file
        """
        i = 0
        filetype = ""
        if context.check_token(i, "INCLUDE") is False:
            return False, 0
        if type(context.scope) is not GlobalScope or context.scope.include_allowed == False:
            context.new_error("INCLUDE_START_FILE", context.peek_token(i))
            return True, i
        val = context.peek_token(i).value.split("include", 1)[1]
        content = Lexer(val, context.peek_token(i).pos[0])
        tkns = content.get_tokens()
        i = 1
        while i < len(tkns) and tkns[i].type in ["TAB", "SPACE"]:
            if tkns[i].type == "TAB":
                context.new_error("TAB_INSTEAD_SPC", tkns[i])
            i += 1
        if i < len(tkns) and tkns[i].type == "LESS_THAN":
            i = len(tkns) - 1
            while i > 0:
                if i < len(tkns) - 1 and tkns[i].type == "DOT":
                    i += 1
                    filetype = tkns[i].value
                    break
                i -= 1
        elif i < len(tkns) and tkns[i].type == "STRING":
            try:
                filetype = tkns[i].value.split(".")[-1][0]
            except:
                filetype = ""
        if filetype and filetype != "h":
            context.new_error("INCLUDE_HEADER_ONLY", context.peek_token(0))
        return False, 0



from norminette.rules import Rule
from norminette.scope import GlobalScope, ControlStructure


class CheckBlockStart(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsBlockStart"]

    def run(self, context):
        """
        Braces signal that the control structure, function, or user defined type can contain
        multiple lines.
        A control structure that has no braces can only contain one instruction line, but can
        contain multiple control structures
        """
        outer = context.scope.get_outer()
        if len(context.history) > 2:
            i = 0
            for item in context.history[::]:
                if item != "IsEmptyLine":
                    if i == 2:
                        hist_1 = item
                    elif i == 3:
                        hist_2 = item
                    i += 1
            if (
                type(context.scope) is GlobalScope
                and context.scope.tmp_scope is not None
                and hist_1 == "IsFuncDeclaration"
                and hist_2 == "IsPreprocessorStatement"
            ):
                context.scope.functions -= 1
                context.scope = context.tmp_scope
                context.scope.multiline = True
                context.tmp_scope = None
        if type(context.scope) == ControlStructure and outer is not None and type(outer) == ControlStructure:
            if outer.multiline == False:
                context.new_error("MULT_IN_SINGLE_INSTR", context.peek_token(0))
        return False, 0



from norminette.rules import PrimaryRule
from norminette.scope import UserDefinedEnum


lbrackets = ["LBRACE", "LPARENTHESIS", "LBRACKET"]
rbrackets = ["RBRACE", "RPARENTHESIS", "RBRACKET"]


class IsEnumVarDecl(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 30
        self.scope = [UserDefinedEnum]

    def assignment_right_side(self, context, pos):
        sep = ["COMMA", "ASSIGN", "NEWLINE"]
        i = context.skip_ws(pos, nl=True)
        while context.peek_token(i) and context.check_token(i, sep) is False:
            if context.check_token(i, lbrackets) is True:
                i = context.skip_nest(i)
            i += 1
        return True, i

    def var_declaration(self, context, pos):
        pclose = ["RPARENTHESIS", "NEWLINE", "SPACE", "TAB"]
        brackets = 0
        parenthesis = 0
        braces = 0
        i = pos
        identifier = False
        while context.peek_token(i) is not None and context.check_token(i, ["COMMA", "RBRACE", "NEWLINE"]) is False:
            if context.check_token(i, "IDENTIFIER") is True and braces == 0 and brackets == 0 and parenthesis == 0:
                identifier = True
            elif context.check_token(i, lbrackets) is True:
                if context.check_token(i, "LBRACE") is True:
                    braces += 1
                if context.check_token(i, "LBRACKET") is True:
                    brackets += 1
                if context.check_token(i, "LPARENTHESIS") is True:
                    parenthesis += 1
            elif context.check_token(i, rbrackets) is True:
                if context.check_token(i, "RBRACE") is True:
                    braces -= 1
                if context.check_token(i, "RBRACKET") is True:
                    brackets -= 1
                if context.check_token(i, "RPARENTHESIS") is True:
                    parenthesis -= 1
            elif context.check_token(i, "ASSIGN") is True:
                if identifier == False:
                    return False, pos
                ret, i = self.assignment_right_side(context, i + 1)
                i -= 1
                if ret is False:
                    return False, pos
            elif context.check_token(i, ["SPACE", "TAB", "MULT", "BWISE_AND", "NEWLINE"]):
                pass
            elif parenthesis == 0 and brackets == 0 and braces == 0:
                return False, 0
            i += 1
        if identifier == False:
            return False, pos
        if context.check_token(i, ["NEWLINE", "COMMA"]) is True:
            return True, i
        return False, pos

    def run(self, context):
        """
        Enum have special var declarations so this catches these specific variables
        """
        ret, i = self.var_declaration(context, 0)
        if ret is False:
            return False, 0
        while ret:
            ret, i = self.var_declaration(context, i)
            if context.check_token(i, ["COMMA"]) is True:
                i += 1
                i = context.eol(i)
                return True, i
            elif context.check_token(i, ["NEWLINE"]) is True:
                i = context.eol(i)
                return True, i
        return False, 0



from norminette.rules import Rule


class CheckEnumVarDecl(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsEnumVarDecl"]

    def run(self, context):
        """
        Checks for nl in declarations
        """
        i = context.skip_ws(0)
        while context.peek_token(i) and context.check_token(i, "COMMA") is False:
            if context.check_token(i, "NEWLINE") is True:
                i += 1
                while context.peek_token(i) is not None and context.check_token(i, "NEWLINE") is False:
                    if context.check_token(i, "LBRACE") is True:
                        return False, 0
                    i += 1
                # context.new_error("NEWLINE_IN_DECL", context.peek_token(i))
                return True, i
            i += 1
        return False, 0



from norminette.rules import PrimaryRule

types = [
    "CHAR",
    "DOUBLE",
    "ENUM",
    "FLOAT",
    "INT",
    "UNION",
    "VOID",
    "LONG",
    "SHORT",
    "SIGNED",
    "UNSIGNED",
    "STRUCT",
    "ENUM",
    "UNION",
    "CONST",
    "REGISTER",
    "STATIC",
    "VOLATILE",
]

op = [
    "MULT",
    "LPARENTHESIS",
    "RPARENTHESIS",
    "LBRACKET",
    "RBRACKET",
    "MINUS",
    "PLUS",
    "DIV",
    "PTR",
    "DOT",
]

ws = ["SPACE", "TAB", "NEWLINE"]


class IsCast(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.primary = True
        self.priority = 15
        self.scope = []

    def run(self, context):
        """
        Catches all casts instructions
        """
        i = 0
        i = context.skip_ws(i, nl=False)
        if context.check_token(i, "LPARENTHESIS") is True:
            ret, i = context.parenthesis_contain(i)
            if ret == "cast":
                while context.peek_token(i) and context.check_token(i, "SEMI_COLON") is False:
                    i += 1
                i += 1
                i = context.skip_ws(i, nl=False)
                i = context.eol(i)
                return True, i
        return False, 0



from norminette.rules import PrimaryRule


class IsLabel(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 10

    def run(self, context):
        """
        Catches label and raises norm error whenever
        """
        i = context.skip_ws(0)
        if context.check_token(i, "IDENTIFIER") is False:
            return False, 0
        i += 1
        i = context.skip_ws(i)
        if context.check_token(i, "COLON") is False:
            return False, 0
        while context.peek_token(i) and context.check_token(i, "NEWLINE") is False:
            i += 1
        i = context.eol(i)
        return True, i



from pathlib import Path

from norminette.lexer import Lexer
from norminette.rules import Rule
from norminette.scope import GlobalScope


class CheckPreprocessorProtection(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsPreprocessorStatement"]

    def run(self, context):
        """
        Header protection must be as follows:
        ```
        #ifndef __FILENAME_H__
        # define __FILENAME_H__
        #endif
        ```
        Any header instruction must be within the header protection
        """
        i = 0
        if type(context.scope) is not GlobalScope:
            return False, 0
        if context.check_token(i, ["IFNDEF", "ENDIF"]) is False or context.filetype != "h":
            return False, 0
        # protection = context.filename.upper().split("/")[-1].replace(".", "_")
        protection = Path(context.filename).name.upper().replace(".", "_")
        val = context.peek_token(i).value.split(" ")[-1]
        content = Lexer(val, context.peek_token(i).pos[0])
        tkns = content.get_tokens()
        if context.check_token(i, "IFNDEF") is True:
            if (
                len(tkns) >= 1
                and tkns[0].value == protection
                and context.scope.header_protection == -1
                and context.preproc_scope_indent == 1
            ):
                if len(context.history) > 1:
                    for i in range(len(context.history) - 2, 0, -1):
                        if context.history[i] != "IsEmptyLine" and context.history[i] != "IsComment":
                            context.new_error("HEADER_PROT_ALL", context.peek_token(0))
                            break
                context.scope.header_protection = 0
            elif len(tkns) < 1 or (tkns[0].value != protection and context.scope.header_protection == -1):
                context.new_error("HEADER_PROT_NAME", context.peek_token(0))
        elif context.check_token(i, "ENDIF") is True:
            if context.scope.header_protection == 1 and context.preproc_scope_indent == 0:
                context.scope.header_protection = 2
        return False, 0



from norminette.rules import Rule
import pdb

assigns = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
]

special_assigns = [
    "INC",
    "DEC"
]


class CheckAssignation(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsAssignation"]

    def check_brace_assign(self, context, i):
        i += 1
        deep = 1
        while context.check_token(i, "RBRACE") is False and deep > 0:
            i += 1
        return True, i

    def check_assign_right(self, context, i, mini_assign=False):
        tmp_typ = None
        start = 0
        full_assign = False
        while context.check_token(i, "SEMI_COLON") is False:
            typ = None
            if context.check_token(i, "LBRACE"):
                ret, i = self.check_brace_assign(context, i)
                if ret == False:
                    return True, i
                break
            if context.check_token(i, "LPARENTHESIS") is True:
                start = i
                tmp_typ, i = context.parenthesis_contain(i)
                if tmp_typ != None:
                    typ = tmp_typ
                if tmp_typ == "assign":
                    context.new_error("MULT_ASSIGN_LINE", context.peek_token(start))
                if tmp_typ is None:
                    tmp = start + 1
                    while context.peek_token(tmp) and context.check_token(tmp, "RPARENTHESIS") is False:
                        if context.check_token(tmp, "COMMA") is True and typ is not None:
                            context.new_error("TOO_MANY_INSTR", context.peek_token(tmp))
                        tmp += 1
            if context.check_token(i, assigns) is True:
                if mini_assign == True:
                    mini_assign = False
                else:
                    context.new_error("MULT_ASSIGN_LINE", context.peek_token(i))
            if context.check_token(i, special_assigns) is True:
                if mini_assign == True:
                    context.new_error("MULT_ASSIGN_LINE", context.peek_token(i))
            i += 1
        return False, 0

    def run(self, context):
        """
        Only one assignation at a time
        Unless the variable is static (or global), you cannot assign its value when you declare it.
        """
        i = 0
        assign_present = False
        mini_assign = False
        while context.check_token(i, "SEMI_COLON") is False:
            if context.check_token(i, assigns + special_assigns) is True and assign_present == False:
                assign_present = True
                if context.check_token(i, special_assigns):
                    mini_assign = True
                return self.check_assign_right(context, i + 1, mini_assign)
            i += 1
        return False, 0



from norminette.lexer.dictionary import brackets
from norminette.lexer.dictionary import keywords
from norminette.lexer.dictionary import operators
from norminette.lexer.dictionary import preproc_keywords


class Token:
    def __init__(self, tkn_type, pos, tkn_value=None):
        self.type = str(tkn_type)
        self.pos = pos
        if tkn_value is not None:
            self.value = str(tkn_value)
            self.length = len(tkn_value)
        else:
            self.value = None
            self.length = 0

    def __repr__(self):
        """
        Token representation for debugging, using the format <TYPE=value>
        or simply <TYPE> when value is None
        """
        r = f"<{self.type}={self.value}>" if self.value else f"<{self.type}>"
        return r

    def test(self):
        return self.__repr__()



from norminette.rules import Rule

whitespaces = ["SPACE", "TAB", "NEWLINE"]

type_specifiers = ["CHAR", "DOUBLE", "ENUM", "FLOAT", "INT", "UNION", "VOID", "SHORT"]

misc_specifiers = ["CONST", "REGISTER", "STATIC", "STRUCT", "VOLATILE"]

size_specifiers = ["LONG", "SHORT"]

sign_specifiers = ["SIGNED", "UNSIGNED"]

arg_separator = ["COMMA", "CLOSING_PARENTHESIS"]


class CheckFuncSpacing(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsFuncDeclaration"]

    def run(self, context):
        """
        Function return type and function name must be separated by a tab
        """
        i = 0
        while i < context.fname_pos:
            if context.check_token(i, "IDENTIFIER") is True and context.peek_token(i).value == "__attribute__":
                # context.new_error("ATTR_EOL", context.peek_token(i))
                break
            i += 1
        i = context.fname_pos - 1
        while context.check_token(i, ["MULT", "BWISE_AND", "LPARENTHESIS"]) is True:
            i -= 1
        if context.peek_token(i).type == "SPACE":
            context.new_error("SPACE_BEFORE_FUNC", context.peek_token(i))
            return False, 0
        if context.peek_token(i).type == "TAB":
            j = i
            while context.peek_token(j).type == "TAB":
                j -= 1
            if j + 1 < i:
                context.new_error("TOO_MANY_TABS_FUNC", context.peek_token(i))
            if context.peek_token(i).type == "SPACE":
                context.new_error("SPACE_BEFORE_FUNC", context.peek_token(i))
        else:
            context.new_error("MISSING_TAB_FUNC", context.peek_token(i))
        return False, 0



from norminette.exceptions import CParsingError
from norminette.lexer.dictionary import operators
from norminette.norm_error import NormError, NormWarning
from norminette.scope import GlobalScope, ControlStructure
from norminette.tools.colors import colors
types = [
    "CHAR",
    "DOUBLE",
    "ENUM",
    "FLOAT",
    "INT",
    "UNION",
    "VOID",
    "LONG",
    "SHORT",
    "SIGNED",
    "UNSIGNED",
]

utypes = ["STRUCT", "ENUM", "UNION"]

glued_operators = [
    "MINUS",
    "PLUS",
    "MULT",
    "DIV",
    "MODULO",
    "TERN_CONDITION",
    "COMMA",
    "GOTO",
    "LABEL",
    "SWITCH",
    "CASE",
    "LESS_OR_EQUAL",
    "GREATER_OR_EQUAL",
    "EQUALS",
    "NOT_EQUAL",
    "LESS_THAN",
    "MORE_THAN",
    "BWISE_XOR",
    "BWISE_OR",
    "BWISE_NOT",
    "BWISE_AND",
    "RIGHT_SHIFT",
    "LEFT_SHIFT",
    "TERN_CONDITION",
    "ASSIGN",
    "AND_ASSIGN",
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
]

operators = [
    "LESS_OR_EQUAL",
    "GREATER_OR_EQUAL",
    "EQUALS",
    "NOT_EQUAL",
    "ASSIGN",
    "AND_ASSIGN",
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "NOT",
    "MINUS",
    "PLUS",
    "MULT",
    "DIV",
    "MODULO",
    "LESS_THAN",
    "MORE_THAN",
    "ELLIPSIS",
    "INC",
    "DEC",
    "PTR",
    "AND",
    "OR",
    "BWISE_XOR",
    "BWISE_OR",
    "BWISE_NOT",
    "BWISE_AND",
    "RIGHT_SHIFT",
    "LEFT_SHIFT",
    "TERN_CONDITION",
    "COMMA",
    "GOTO",
    "LABEL",
    "SWITCH",
    "CASE",
]
misc_specifiers = [
    "CONST",
    "RESTRICT",
    "REGISTER",
    "STATIC",
    "VOLATILE",
    "EXTERN",
    "INLINE",
    "RESTRICT",
    "SIGNED",
    "UNSIGNED",
]

assigns = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
]

size_specifiers = ["LONG", "SHORT"]

sign_specifiers = ["SIGNED", "UNSIGNED"]

whitespaces = ["SPACE", "TAB", "ESCAPED_NEWLINE", "NEWLINE"]

arg_separator = ["COMMA", "CLOSING_PARENTHESIS"]

import pdb

class Context:
    def __init__(self, filename, tokens, debug=0, added_value=[]):
        #Header relative informations
        self.header_started = False
        self.header_parsed = False
        self.header = ""
        # File relative informations
        self.filename = filename
        self.filetype = filename.split(".")[-1]  # ?
        self.tokens = tokens
        self.debug = int(debug)

        # Rule relative informations
        self.history = []
        self.errors = []
        self.warnings = []
        self.tkn_scope = len(tokens)

        # Scope informations
        self.scope = GlobalScope()
        self.func_alignment = 0
        self.sub = None
        self.fname_pos = 0
        self.arg_pos = [0, 0]

        # Preprocessor handling
        self.preproc_scope_indent = 0
        self.skip_define_error = True if added_value is not None and "CheckDefine" in added_value else False

    def peek_token(self, pos):
        return self.tokens[pos] if pos < len(self.tokens) else None

    def pop_tokens(self, stop):
        self.tokens = self.tokens[stop:]

    def check_token(self, pos, value):
        """Compares the token at 'pos' against a value or list of values"""
        tkn = self.peek_token(pos)

        if tkn is None:
            return None

        if isinstance(value, list):
            if tkn.type in value:
                return True
            return False

        return tkn.type == value

    def find_in_scope(self, value, nested=True):
        nests = 0
        for i in range(0, self.tkn_scope):
            tkn = self.peek_token(i)
            if self.check_token(i, ["LBRACKET", "LPARENTHESIS", "LBRACE"]) is True:
                nests += 1
            if self.check_token(i, ["RBRACKET", "RPARENTHESIS", "RBRACE"]) is True:
                nests -= 1
            if tkn.type == value and (nested == True or (nests == 0 and nested == False)):
                return i
        return -1

    def new_error(self, errno, tkn):
        self.errors.append(NormError(errno, tkn.pos[0], tkn.pos[1]))

    def new_warning(self, errno, tkn):
        self.warnings.append(NormWarning(errno, tkn.pos[0], tkn.pos[1]))

    def get_parent_rule(self):
        if len(self.history) == 0:
            return ""
        return self.history[-1 if len(self.history) == 1 else -2]

    def update(self):
        """Updates informations about the context and  the scope if needed
        after a primary rule has succeeded.
        Do nothing on empty lines since they can be anywhere
        """
        if len(self.history) > 0 and (
            self.history[-1] == "IsEmptyLine"
            or self.history[-1] == "IsComment"
            or self.history[-1] == "IsPreprocessorStatement"
        ):
            return
        if self.sub is not None:
            self.scope = self.sub
            self.sub = None
        if type(self.scope) is ControlStructure and self.scope.multiline is False and self.scope.instructions > 0:
            self.scope = self.scope.outer()
            self.sub = None
            self.update()
        self.arg_pos = [0, 0]

    def dprint(self, rule, pos):
        """Debug printing, shows the primary rules that succeed in matching
        tokens and print the matching tokens
        """
        if self.debug < 2:
            return
        print(
            f"{colors(self.filename, 'cyan')} - {colors(rule, 'green')} \
In \"{self.scope.name}\" from \
\"{self.scope.parent.name if self.scope.parent is not  None else None}\" line {self.tokens[0].pos[0]}\":"
        )
        i = 0
        for t in self.tokens[:pos]:
            if i == 0:
                print("\t\t", end="")
            if t.type == "NEWLINE":
                print(t)
                i = 0
            else:
                print(t, end=" ")
                i += 1
        if pos - 1 < len(self.tokens) and self.tokens[pos - 1].type != "NEWLINE":
            print("")
        elif len(self.tokens) == 1 and self.tokens[-1].type != "NEWLINE":
            print ("")

    def eol(self, pos):
        """Skips white space characters (tab, space) until end of line
        (included) or any other token (excluded)
        """
        while self.check_token(pos, ["TAB", "SPACE", "NEWLINE"]) is True:
            if self.check_token(pos, "NEWLINE"):
                pos += 1
                break
            pos += 1
        return pos

    def skip_ws(self, pos, nl=False):
        ws = whitespaces[:]
        if nl == False:
            ws.remove("NEWLINE")
        while self.check_token(pos, ws):
            pos += 1
        return pos

    def skip_nest_reverse(self, pos):
        """Skips anything between two brackets, parentheses or braces starting
        at 'pos', if the brackets, parentheses or braces are not closed or
        are closed in the wrong order an error shall be raised
        """
        rbrackets = ["LBRACKET", "LBRACE", "LPARENTHESIS"]
        lbrackets = ["RBRACKET", "RBRACE", "RPARENTHESIS"]
        try:
            c = self.peek_token(pos).type
        except:
            raise CParsingError(f"Error: Unexpected EOF line {pos}")
        if c not in lbrackets:
            return pos
        c = rbrackets[lbrackets.index(c)]
        i = pos - 1
        while self.peek_token(i) is not None:
            if self.check_token(i, lbrackets) is True:
                i = self.skip_nest_reverse(i)
                if i == -1:
                    return -1
            elif self.check_token(i, rbrackets) is True:
                if c == self.peek_token(i).type:
                    return i
            i -= 1
        raise CParsingError(
            "Error: Nested parentheses, braces or brackets\
 are not correctly closed"
        )

        return -1

    def skip_nest(self, pos):
        """Skips anything between two brackets, parentheses or braces starting
        at 'pos', if the brackets, parentheses or braces are not closed or
        are closed in the wrong order an error shall be raised
        """
        lbrackets = ["LBRACKET", "LBRACE", "LPARENTHESIS"]
        rbrackets = ["RBRACKET", "RBRACE", "RPARENTHESIS"]
        try:
            c = self.peek_token(pos).type
        except:
            raise CParsingError(f"Error: Code ended unexpectedly.")
        if c not in lbrackets:
            return pos
        c = rbrackets[lbrackets.index(c)]
        i = pos + 1
        while self.peek_token(i) is not None:
            if self.check_token(i, lbrackets) is True:
                i = self.skip_nest(i)
                if i == -1:
                    return -1
            elif self.check_token(i, rbrackets) is True:
                if c == self.peek_token(i).type:
                    return i
            i += 1
        raise CParsingError(
            "Error: Nested parentheses, braces or brackets\
 are not correctly closed"
        )

        return -1

    def skip_misc_specifier(self, pos, nl=False):
        i = self.skip_ws(pos, nl=nl)
        if self.check_token(i, "IDENTIFIER"):
            tmp = self.skip_misc_specifier(i + 1)
            if tmp != i + 1:
                tmp = i
        if self.check_token(i, "MULT"):
            tmp = i + 1
            while self.check_token(tmp, "MULT"):
                tmp += 1
            tmp = self.skip_ws(tmp, nl=nl)
            if self.check_token(tmp, misc_specifiers):
                i = tmp
                i = self.skip_ws(i, nl=nl)
        while self.check_token(i, misc_specifiers):
            i += 1
            i = self.skip_ws(i, nl=nl)
            if self.check_token(i, "MULT"):
                tmp = i + 1
                tmp = self.skip_ws(i, nl=nl)
                if self.check_token(tmp, misc_specifiers):
                    i = tmp
        return i

    def skip_typedef(self, pos):
        i = self.skip_ws(pos)
        if self.check_token(i, "TYPEDEF"):
            i += 1
            i += self.skip_misc_specifier(pos)
        return i

    def check_type_specifier(self, pos, user_def_type=False, nl=False):
        """Returns (True, pos + n) if the tokens from 'pos' to 'n' could match
        a valid type specifier. Valid type specifiers consist of:
            -an optionnal 'misc' specifier (const, register, volatile ...)
            -an optionnal size specifier (long or short)
            -a type specifier (int, char, double, etc...)
                OR an IDENTIFIER
                OR a user type specifier (struct, union, enum) + IDENTIFIER
        """
        i = self.skip_misc_specifier(pos, nl=nl)
        i = self.skip_ws(i, nl=nl)
        if user_def_type is True:
            if self.check_token(i, utypes + ["TYPEDEF"]) is True:
                while self.check_token(i, whitespaces + utypes + ["TYPEDEF"]) is True:
                    i += 1
                if self.check_token(i, "IDENTIFIER") is True:
                    i += 1
                    return True, i
                # Raise CParsingError?
            if self.check_token(i, types + ["IDENTIFIER", "TYPEDEF"]) is False:
                return False, 0
            if self.check_token(i, "IDENTIFIER") is True:
                i += 1
                return True, i
            while self.check_token(i, types + whitespaces + ["TYPEDEF"]) is True:
                i += 1
            return True, i
        else:
            if self.check_token(i, utypes) is True:
                i += 1
                i = self.skip_ws(i)
                if self.check_token(i, "IDENTIFIER") is True:
                    i += 1
                    return True, i
            if self.check_token(i, types + ["IDENTIFIER"]) is False:
                return False, 0
            if self.check_token(i, "IDENTIFIER") is True:
                i += 1
                #i = self.skip_ws(i)
                return True, i + 1
            while self.check_token(i, types + whitespaces + ["MULT", "BWISE_AND"]) is True:
                i += 1
            tmp = self.skip_misc_specifier(i, nl=nl)
            if tmp == i:
                return True, i - 1
            else:
                return True, tmp

    def check_identifier(self, pos, nl=False):
        """
        Determines the function return value or the variable type and returns
        an iterator to the next token
        """
        i = pos
        p = 0
        i = self.skip_misc_specifier(i, nl=nl)
        while self.check_token(i, whitespaces + ["MULT", "LPARENTHESIS"]) is True:
            if self.check_token(i, "LPARENTHESIS"):
                p += 1
            if self.check_token(i, "MULT") and self.check_token(i + 1, "CONST"):
                i += 1
            i += 1
        i = self.skip_misc_specifier(i, nl=nl)
        if self.check_token(i, "IDENTIFIER"):
            while p and self.check_token(i, whitespaces + ["RPARENTHESIS"]) is True:
                if self.check_token(i, "RPARENTHESIS"):
                    p -= 1
                i += 1
            return True, i
        return False, pos

    def is_glued_operator(self, pos):
        """
        Returns True if operator (among +-) at given pos is glued to identifier, number
        or constant
        """
        glued = [
            "LPARENTHESIS",
            "LBRACKET",
            "LBRACE",
        ]
        glued = glued + glued_operators
        start = pos
        if (
            self.check_token(
                pos,
                ["PLUS", "MINUS", "BWISE_OR", "BWISE_AND", "BWISE_NOT", "BWISE_XOR"],
            )
            is False
        ):
            return False
        pos += 1
        pos = self.skip_ws(pos, nl=False)
        if self.check_token(pos, ["IDENTIFIER", "CONSTANT", "MULT", "BWISE_AND"]) is False:
            return False
        pos = start - 1
        while (self.check_token(pos, ["SPACE", "TAB"])) is True:
            pos -= 1
        if self.check_token(pos, glued) is True:
            return True
        return False

    def is_operator(self, pos):
        """
        Returns True if the given operator (among '*&') is an actual operator,
        and returns False if said operator is a pointer/adress indicator
        """
        i = 0
        start = pos + 1
        pos -= 1
        if self.history[-1] == "IsFuncPrototype" or self.history[-1] == "IsFuncDeclaration":
            return False
        if self.check_token(start, ["RPARENTHESIS", "MULT"]) is True:
            return False
        start = self.skip_ws(start, nl=False)
        if self.check_token(start, ["SIZEOF"]) is True:
            return True
        if self.history[-1] == "IsVarDeclaration":
            bracketed = False
            tmp = pos
            right_side = False
            while tmp > 0:
                if self.check_token(tmp, ["RBRACKET", "RPARENTHESIS"]) is True:
                    tmp = self.skip_nest_reverse(tmp) - 1
                if self.check_token(tmp, ["ASSIGN"]) is True:
                    right_side = True
                if self.check_token(tmp, "LBRACKET") is True:
                    bracketed = True
                tmp -= 1
            if right_side == False and bracketed == False:
                return False
        skip = 0
        value_before = False
        while pos > 0:
            if self.check_token(pos, ["RBRACKET", "RPARENTHESIS"]) is True:
                value_before = True
                pos = self.skip_nest_reverse(pos) - 1
                if self.check_token(pos + 1, "LPARENTHESIS") is True and self.parenthesis_contain(pos + 1)[0] == "variable":
                    return True
                if self.check_token(pos + 1, "LPARENTHESIS") is True and self.parenthesis_contain(pos + 1)[0] == "cast":
                    return False
                skip = 1
            if self.check_token(pos, ["IDENTIFIER", "CONSTANT", "SIZEOF", "CHAR_CONST"]) is True:
                if self.check_token(pos, "IDENTIFIER") is True and self.check_token(pos + 1, "TAB") is True:
                    return False
                return True
            if self.check_token(pos, ["COMMA", "LPARENTHESIS", "LBRACKET"] + operators) is True and skip == 1 and self.parenthesis_contain(pos + 1)[0] != "cast":
                return True
            if self.check_token(pos, ["LBRACKET", "LPARENTHESIS", "MULT", "BWISE_AND", "COMMA"] + operators + types):
                return False
            pos -= 1
        if value_before == True:
            return True
        else:
            return False

    def parenthesis_contain(self, i, ret_store=None):
        """
        Explore parenthesis to return its content
        Function, pointer, cast, or other
        Uses basic string as return value and skips to the end of the parenthesis nest
        """
        if self.check_token(i, "LPARENTHESIS") is False:
            return None, i
        start = i
        ws = ["SPACE", "TAB", "NEWLINE"]
        i += 1
        deep = 1
        nested_id = False
        identifier = None
        pointer = None
        sizeof = False
        id_only = True
        if self.check_token(start - 1, "SIZEOF") is True:
            sizeof = True
        i = self.skip_ws(i)
        while deep > 0 and self.peek_token(i) is not None:
            #print (self.peek_token(i), deep, identifier, self.check_token(i, "NULL"))
            if self.check_token(i, "RPARENTHESIS"):
                deep -= 1
            elif self.check_token(i, "LPARENTHESIS"):
                deep += 1
                #if identifier is not None and deep >= 0:
                    #return "pointer", self.skip_nest(start)
            elif deep > 1 and identifier == True and self.check_token(i, ["NULL", "IDENTIFIER"]):
                return "fct_call", self.skip_nest(start)
            elif self.check_token(i, "COMMA") and nested_id == True:
                return "function", self.skip_nest(start)
            elif self.check_token(i, assigns) and deep == 1:
                return "assign", self.skip_nest(start)
            elif self.check_token(i, "PTR") and deep == 1:
                return "variable", self.skip_nest(start)
            elif self.check_token(i, "COMMA"):
                return None, self.skip_nest(start)
            elif self.check_token(i, ws):
                pass
            elif self.check_token(i, types):
                tmp = start - 1
                while self.check_token(tmp, ["SPACE", "TAB"]) == True:
                    tmp -= 1
                if self.check_token(tmp, "SIZEOF") == True:
                    return None, self.skip_nest(start)
                tmp = start + 1
                while self.check_token(tmp, "RPARENTHESIS") is False:
                    if self.check_token(tmp, ["LPARENTHESIS", "IDENTIFIER"]) is True:
                        return None, self.skip_nest(start)
                    tmp += 1
                if deep == 1:
                    return "cast", self.skip_nest(start)
            elif self.check_token(i, "IDENTIFIER"):
                tmp = i + 1
                if (identifier is not True and pointer == True) or ret_store is not None:
                    nested_id = True
                if identifier is not True and self.check_token(tmp, "RPARENTHESIS") and self.scope.name == "Function" and deep == 1 and pointer == None and sizeof == False:
                    tmp = self.skip_nest(start) + 1
                    tmp = self.skip_ws(tmp)
                    if self.check_token(tmp, ["IDENTIFIER", "CONSTANT", "MINUS", "PLUS"]) is False:
                        return None, self.skip_nest(start)
                    return "cast", self.skip_nest(start)
                identifier = True
                tmp = self.skip_ws(tmp)
                if pointer == True:
                    if self.check_token(tmp, "LBRACKET"):
                        tmp = self.skip_nest(tmp)
                        tmp += 1
                    while self.check_token(tmp, "RPARENTHESIS"):
                        tmp += 1
                        # start = tmp
                    tmp = self.skip_ws(tmp)
                    if self.check_token(tmp, "LPARENTHESIS"):
                        return "pointer", self.skip_nest(start)
                    elif self.check_token(tmp, "RPARENTHESIS"):
                        return None, self.skip_nest(start)
            elif self.check_token(i, ["MULT", "BWISE_AND"]):
                tmp = i + 1
                pointer = True
                if identifier != None:
                    tmp = start - 1
                    while self.check_token(tmp, ["SPACE", "TAB"]) == True:
                        tmp -= 1
                    if self.check_token(tmp, "SIZEOF") == True:
                        return None, self.skip_nest(start)
                    tmp = self.skip_ws(i + 1)
                    if self.check_token(tmp, "RPARENTHESIS") is True:
                        return "cast", self.skip_nest(start)
            i += 1
        if identifier == True and id_only == True:
            return "var", self.skip_nest(start)
        return None, self.skip_nest(start)



from norminette.rules import Rule
from norminette.scope import GlobalScope
import pdb

class CheckNewlineIndent(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsDeclaration", "IsAssignation", "IsCast", "IsExpressionStatement"]

    def run(self, context):
        """
        If a line has a newline inside, we must check for indent - authorized : same indent/same + 1 indent
        """
        if context.scope.name != "Function":
            return False, 0
        expected = context.scope.indent
        i = context.find_in_scope("NEWLINE", nested=False) + 1
        if i != -1 and i < context.tkn_scope - 2:
            start = i
            got = 0
            while context.check_token(start + got, "TAB"):
                got += 1
            if got > expected + 1:
                context.new_error("TOO_MANY_TAB", context.peek_token(start))
            if got < expected:
                context.new_error("TOO_FEW_TAB", context.peek_token(start))
        return False, 0



from norminette.rules import Rule

types = [
    "INT",
    "VOID",
    "CHAR",
    "LONG",
    "DOUBLE",
    "SHORT",
    "SIGNED",
    "UNSIGNED",
    "STATIC",
    "CONST",
    "IDENTIFIER",
    "ENUM",
    "STRUCT",
    "UNION",
    "VOLATILE",
    "EXTERN",
    "SPACE",
    "TAB",
]


class CheckGlobalNaming(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsVarDeclaration"]

    def run(self, context):
        """
        Global variable names must be preceded by g_
        """
        i = 0
        last_id = ""
        if context.scope.name != "GlobalScope":
            return False, 0
        i = context.skip_ws(i)
        _, i = context.check_type_specifier(i)
        i = context.skip_ws(i)
        while context.check_token(i, "IDENTIFIER") is False:
            i += 1
        if context.peek_token(i) is not None and context.peek_token(i).value != "environ":
            context.new_warning("GLOBAL_VAR_DETECTED", context.peek_token(0))
            if context.peek_token(i).value.startswith("g_") is False:
                context.new_error("GLOBAL_VAR_NAMING", context.peek_token(i))
        return False, i



from norminette.rules import Rule


allowed_in_header = [
    "IsVarDeclaration",
    "IsUserDefinedType",
    "IsPreprocessorStatement",
    "IsEmptyLine",
    "IsBlockStart",
    "IsBlockEnd",
    "IsComment",
    "IsEndOfLine",
    "IsFuncPrototype",
]

must_be_within_define = [
    "IsVarDeclaration",
    "IsUserDefinedType",
    "IsFuncPrototype",
]


class CheckInHeader(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = [
            "IsVarDeclaration",
            "IsUserDefinedType",
            "IsPreprocessorStatement",
            "IsEmptyLine",
            "IsBlockStart",
            "IsBlockEnd",
            "IsComment",
            "IsEndOfLine",
            "IsFuncPrototype",
        ]

    def run(self, context):
        """
        Each .h file must be protected against double inclusion
        Instructions allowed in header files:
            - Variable declaration
            - User defined types
            - Comments
            - Function prototypes
        """
        if context.filetype != "h":
            return False, 0
        sc = context.scope
        while sc.name != "GlobalScope":
            sc = sc.get_outer()
        if context.history[-1] not in allowed_in_header:
            context.new_error("FORBIDDEN_IN_HEADER", context.peek_token(0))
            return False, 0
        elif context.history[-1] in must_be_within_define and sc.header_protection != 1:
            context.new_error("HEADER_PROT_ALL", context.peek_token(0))
        return False, 0



from norminette.rules import Rule

types = ["INT", "FLOAT", "CHAR", "DOUBLE", "LONG", "SHORT"]


class CheckFuncDeclaration(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = ["IsFuncDeclaration", "IsFuncPrototype"]

    def run(self, context):
        """
        Maximum 4 arguments in a function
        Function declaration must be preceded by a newline
        """
        i = 0
        tmp = 0
        start = 0
        arg = 1
        while context.check_token(tmp, ["SEMI_COLON", "NEWLINE"]) is False:
            if context.check_token(tmp, "LBRACE") is True:
                context.new_error("BRACE_NEWLINE", context.peek_token(tmp))
            tmp += 1
        # if tmp < context.tkn_scope - 2:
        # context.new_error("NEWLINE_IN_DECL", context.peek_token(tmp))
        # this is a func declaration
        if context.history[-1] == 'IsFuncDeclaration':
#        if context.check_token(tmp, "SEMI_COLON") is False:
            i = 2
            length = len(context.history)
            while length - i >= 0 and (context.history[-i] == "IsPreprocessorStatement"
                or context.history[-i] == "IsComment"
                or context.history[-i] == "IsFuncDeclaration"
            ):
                i += 1
            if length - i > 0 and context.history[-i] != "IsEmptyLine":
                context.new_error("NEWLINE_PRECEDES_FUNC", context.peek_token(start))
        i = context.fname_pos + 1
        while (context.check_token(i, ["RPARENTHESIS"])) is True: #, "SPACE", "TAB"])) is True:
            i += 1
        if context.check_token(i, "LPARENTHESIS") is False:
            context.new_error("EXP_PARENTHESIS", context.peek_token(i))
        i += 1
        deep = 1
        while deep > 0:
            if context.check_token(i, "LPARENTHESIS"):
                i = context.skip_nest(i)
            if context.check_token(i, "RPARENTHESIS"):
                deep -= 1
            if context.check_token(i, "COMMA"):
                arg += 1
            i += 1
        if context.check_token(i - 1, ["SPACE", "TAB"]) is True:
            tmp = i - 1
            while context.check_token(tmp, ["SPACE", "TAB"]) is True:
                tmp -= 1
            if context.check_token(tmp, "NEWLINE") is False:
                context.new_error("NO_SPC_BFR_PAR", context.peek_token(i))
        if arg > 4:
            context.new_error("TOO_MANY_ARGS", context.peek_token(i))
        arg = []
        while context.check_token(i, ["NEWLINE", "SEMI_COLON"]) is False:
            i += 1
        if context.check_token(i - 1, ["TAB", "SPACE"]):
            context.new_error("SPC_BEFORE_NL", context.peek_token(i))
        return False, 0



import pdb

from norminette.rules import Rule
from norminette.scope import GlobalScope


class CheckLineIndent(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Each new scope (function, control structure, struct/enum type declaration) adds a tab to the general indentation
        """
        expected = context.scope.indent 
        if (
            context.history[-1] == "IsEmptyLine"
            or context.history[-1] == "IsComment"
            or context.history[-1] == "IsPreprocessorStatement"
        ):
            return False, 0
        if (
            context.history[-1] != "IsPreprocessorStatement"
            and type(context.scope) is GlobalScope
            and context.scope.include_allowed == True
        ):
            context.scope.include_allowed = False
        got = 0
        while context.check_token(got, "TAB"):
            got += 1
        if context.check_token(got, ["LBRACE", "RBRACE"]) and expected > 0:
            if context.check_token(got, "RBRACE") is True:
                expected -= 1
            else:
                hist = context.history[: len(context.history) - 1]
                for item in hist[::-1]:
                    if item == "IsEmptyLine" or item == "IsComment" or item == "IsPreprocessorStatement":
                        continue
                    if item not in [
                        "IsControlStatement",
                        "IsFuncDeclaration",
                        "IsUserDefinedType",
                    ]:
                        break
                    else:
                        expected -= 1
                    break
        if expected > got:
            context.new_error("TOO_FEW_TAB", context.peek_token(0))
            return False, got
        elif got > expected:
            context.new_error("TOO_MANY_TAB", context.peek_token(0))
            return False, got
        return False, 0



from norminette.rules import Rule


class CheckLineLen(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Lines must not be over 80 characters long
        """
        i = 0
        line_too_long = {}
        for tkn in context.tokens[: context.tkn_scope]:
            if tkn.pos[1] > 81 and tkn.pos[0] not in line_too_long:
                context.new_error("LINE_TOO_LONG", tkn)
                line_too_long[tkn.pos[0]] = True
            i += 1
        return False, 0



from norminette.rules import Rule


class CheckGeneralSpacing(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = [
            "IsDeclaration",
            "IsControlStatement",
            "IsExpressionStatement",
            "IsAssignation",
            "IsFunctionCall",
        ]

    def run(self, context):
        """
        Checks for tab/space consistency
        """
        if context.scope.name == "UserDefinedType":
            return False, 0
        i = context.skip_ws(0)
        while i < context.tkn_scope:
            if context.check_token(i, "TAB") is True:
                context.new_error("TAB_INSTEAD_SPC", context.peek_token(i))
                break
            if context.check_token(i, ["NEWLINE", "ESCAPED_NEWLINE"]) is True:
                i = context.skip_ws(i + 1, nl=True)
            i += 1
        return False, 0



from norminette.context import GlobalScope
from norminette.rules import Rule


class CheckLineCount(Rule):
    def __init__(self):
        super().__init__()
        self.depends_on = []

    def run(self, context):
        """
        Each function can only have 25 lines between its opening and closing brackets
        """
        for t in context.tokens[: context.tkn_scope]:
            if t.type == "NEWLINE" or t.type == "ESCAPED_NEWLINE":
                context.scope.lines += 1

        if type(context.scope) is GlobalScope:

            if context.get_parent_rule() == "CheckFuncDeclarations" and context.scope.lines > 25:
                context.new_error("TOO_MANY_LINES", context.tokens[context.tkn_scope])
            return False, 0

        if context.get_parent_rule() == "CheckBrace":
            if "LBRACE" in [t.type for t in context.tokens[: context.tkn_scope + 1]]:
                if type(context.scope) is GlobalScope:
                    return False, 0
            else:
                if context.scope.lvl == 0:
                    return False, 0

        return False, 0



from norminette.rules import PrimaryRule


class IsDeclaration(PrimaryRule):
    def __init__(self):
        super().__init__()
        self.priority = 5
        self.scope = []

    def run(self, context):
        # return False, 0
        i = context.skip_ws(0, nl=False)
        p = 0
        ident = None
        while context.peek_token(i) is not None and context.check_token(i, "SEMI_COLON") is False:
            if context.check_token(i, "LPARENTHESIS"):
                p += 1
            if context.check_token(i, "RPARENTHESIS"):
                p -= 1
            if context.check_token(i, ["IDENTIFIER", "NULL"]):
                ident = context.peek_token(i)
            i += 1
        i += 1
        i = context.skip_ws(i, nl=False)
        if context.check_token(i, "NEWLINE"):
            i += 1
        if p == 0 and ident != None:
            return True, i
        return False, 0
